from enum import Enum


SIDE_BUY = 'BUY'
SIDE_SELL = 'SELL'

ORDER_TYPE_LIMIT = 'LIMIT'
ORDER_TYPE_MARKET = 'MARKET'
ORDER_TYPE_STOP_LOSS = 'STOP_LOSS'
ORDER_TYPE_STOP_LOSS_LIMIT = 'STOP_LOSS_LIMIT'
ORDER_TYPE_TAKE_PROFIT = 'TAKE_PROFIT'
ORDER_TYPE_TAKE_PROFIT_LIMIT = 'TAKE_PROFIT_LIMIT'
ORDER_TYPE_LIMIT_MAKER = 'LIMIT_MAKER'


class Resolution(Enum):
    ONE_HOUR = '60'
    THREE_HOURS = '180'
    SIX_HOURS = '360'
    TWELVE_HOURS = '720'
    ONE_DAY = 'D'
    TWO_DAYS = '2D'
    THREE_DAYS = '3D'
