__name__ = "wallex"
__version__ = "0.4.5"
__author__ = "amiwrpremium"
__reason__ = 'OK'
__email__ = "amiwrpremium@gmail.com"
__license__ = "MIT"
__url__ = "https://github.com/amiwrpremium/wallex"
__description__ = "Wallex cryptocurrency exchange python SDK"
